from __future__ import annotations

import contextlib
import dataclasses
import threading
import typing

from ._metric_base import (
    _to_timestamp,
    DynamicTimestamp,
    MetricFamily,
    MetricPoint,
    MetricValueSet,
)
from ._path import LabelsData


@contextlib.contextmanager
def opt_lock(lock: threading.Lock | None) -> typing.Iterator[None]:
    if lock:
        with lock:
            yield
    else:
        yield


@dataclasses.dataclass(slots=True, kw_only=True)
class _NumberValue:
    value: float = 0
    timestamp: int | None = None

    def _get_value_set(self) -> MetricValueSet:
        return MetricValueSet(
            values=[("", self.value)],
            timestamp=self.timestamp,
        )


class _SimpleNumber(MetricPoint):
    __slots__ = ('_inc_lock', '_value')

    def __init__(self, *, value: float, inc_lock: threading.Lock | None) -> None:
        super().__init__()
        self._inc_lock = inc_lock
        self._value = _NumberValue(value=value)

    def _get_value_set(self) -> MetricValueSet:
        return self._value._get_value_set()

    def set(self, value: float, *, timestamp: DynamicTimestamp = None) -> None:
        self._value = _NumberValue(value=value, timestamp=_to_timestamp(timestamp))

    def inc_no_lock(self, add: float = 1, *, timestamp: DynamicTimestamp = None) -> None:
        self._value = _NumberValue(value=self._value.value + add, timestamp=_to_timestamp(timestamp))

    def inc(self, add: float = 1, *, timestamp: DynamicTimestamp = None) -> None:
        timestamp = _to_timestamp(timestamp)
        if self._inc_lock:
            with self._inc_lock:
                self.inc_no_lock(add=add, timestamp=timestamp)
        else:
            self.inc_no_lock(add=add, timestamp=timestamp)


class Counter(_SimpleNumber):
    __slots__ = ()

    def __init__(self, *, value: int, inc_lock: threading.Lock | None) -> None:
        super().__init__(value=value, inc_lock=inc_lock)


class CounterFamily(MetricFamily[Counter]):
    __slots__ = ()

    TYPE: typing.ClassVar[str] = 'counter'
    SUFFIXES: typing.ClassVar[list[str]] = ['', '_created']

    DEFAULT_INC_LOCK: typing.ClassVar[threading.Lock] = threading.Lock()

    def create(self, *, value: int = 0, labels: LabelsData) -> Counter:
        point = Counter(value=value, inc_lock=self.DEFAULT_INC_LOCK)
        return self._init_point(labels, point)


class Gauge(_SimpleNumber):
    __slots__ = ()

    def __init__(self, *, value: float, inc_lock: threading.Lock | None) -> None:
        super().__init__(value=value, inc_lock=inc_lock)


class GaugeFamily(MetricFamily[Gauge]):
    __slots__ = ()

    TYPE: typing.ClassVar[str] = 'gauge'
    SUFFIXES: typing.ClassVar[list[str]] = ['']

    DEFAULT_INC_LOCK: typing.ClassVar[threading.Lock] = threading.Lock()

    def create(self, *, value: float = 0, labels: LabelsData) -> Gauge:
        point = Gauge(value=value, inc_lock=self.DEFAULT_INC_LOCK)
        return self._init_point(labels, point)


class Summary(MetricPoint):
    # TODO: quantile support
    __slots__ = ('_lock', '_sum', '_count', '_timestamp')

    def __init__(self, *, lock: threading.Lock | None = None) -> None:
        super().__init__()
        self._lock = lock
        self._sum: float = 0
        self._count: int = 0
        self._timestamp: int | None = 0

    def _get_value_set(self) -> MetricValueSet:
        with opt_lock(self._lock):
            return MetricValueSet(
                values=[
                    ("_sum", self._sum),
                    ("_count", self._count),
                ],
                timestamp=self._timestamp,
            )

    def observe(self, value: float, timestamp: DynamicTimestamp = None) -> None:
        timestamp = _to_timestamp(timestamp)
        with opt_lock(self._lock):
            self._sum += value
            self._count += 1
            self._timestamp = timestamp


class SummaryFamily(MetricFamily[Summary]):
    __slots__ = ()

    TYPE: typing.ClassVar[str] = 'summary'
    SUFFIXES: typing.ClassVar[list[str]] = ['', '_count', '_sum', '_created']
    RESERVED_LABELS: typing.ClassVar[list[str]] = ['quantile']

    DEFAULT_LOCK: typing.ClassVar[threading.Lock] = threading.Lock()

    def create(self, *, value: float = float('nan'), labels: LabelsData) -> Summary:
        point = Summary(lock=self.DEFAULT_LOCK)
        return self._init_point(labels, point)
