from __future__ import annotations

import math
import threading
import typing
import weakref

from ._metric_base import (
    MetricFamily,
    MetricPoint,
    MetricValueSet,
)


_INF = float("inf")
_MINUS_INF = float("-inf")


def _floatToGoString(d: float) -> str:
    if d == _INF:
        return '+Inf'
    elif d == _MINUS_INF:
        return '-Inf'
    elif math.isnan(d):
        return 'NaN'
    else:
        s = repr(d)
        dot = s.find('.')
        # Go switches to exponents sooner than Python.
        # We only need to care about positive values for le/quantile.
        if d > 0 and dot > 6:
            mantissa = '{0}.{1}{2}'.format(
                s[0], s[1:dot], s[dot + 1:]).rstrip('0.')
            return '{0}e+0{1}'.format(mantissa, dot - 1)
        return s


class _FamilyInstance:
    def __init__(self, family: MetricFamily) -> None:
        self._family = family
        self._lock = threading.Lock()
        self._points: weakref.WeakValueDictionary[str, MetricPoint] = weakref.WeakValueDictionary()

    def _add_point(self, point: MetricPoint) -> None:
        key = point.labels.key
        if not point.family is self._family:
            raise KeyError(f"Point {point.family.name}{key} has different family {point.family} than {self._family}")
        with self._lock:
            if key in self._points:
                raise KeyError(f"Point {key} already exists in family instance {self._family.name}")
            self._points[key] = point

    def _get_value_sets(self) -> list[tuple[str, MetricValueSet]]:
        return [
            (point.labels.key, point._get_value_set())
            for point in self._points.values()
        ]


class MetricGroup(typing.Protocol):
    def metric_points(self) -> typing.Iterable[MetricPoint]:
        ...


class Registry:
    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._families: dict[str, _FamilyInstance] = {}
        self._names: dict[str, _FamilyInstance] = {}

    def _register_family(self, family: MetricFamily) -> _FamilyInstance:
        with self._lock:
            if family.name in self._families:
                instance = self._families[family.name]
                if instance._family is family:
                    return instance
                raise KeyError(f"Metric family {family.name} already reserved")
            if family.name in self._names:
                raise KeyError(f"Metric family {family.name} already reserved by {self._names[family.name]._family}")
            for suffix in family.SUFFIXES:
                full_name = family.name + suffix
                if full_name in self._families:
                    raise KeyError(f"Metric name {full_name} already reserved by other family")
                if full_name in self._names:
                    raise KeyError(f"Metric name {full_name} already reserved by {self._names[full_name]._family}")
            # now register
            instance = _FamilyInstance(family=family)
            self._families[family.name] = instance
            for suffix in family.SUFFIXES:
                full_name = family.name + suffix
                self._names[full_name] = instance
            return instance

    def register(self, point: MetricPoint) -> None:
        instance = self._register_family(point.family)
        instance._add_point(point)

    def register_group(self, group: MetricGroup) -> None:
        for point in group.metric_points():
            self.register(point)

    def collect(self) -> str:
        from io import StringIO
        result = StringIO()

        def escape_text(text: str) -> str:
            return text.replace('\\', r'\\').replace('\n', r'\n')

        with self._lock:
            families = list(self._families.values())

        first = True
        for fam_instance in families:
            if first:
                first = False
            else:
                result.write("\n")
            family = fam_instance._family
            if family.help:
                result.write(f"# HELP {family.name} {escape_text(family.help)}\n")
            if family.unit:
                result.write(f"# UNIT {family.name} {family.unit}\n")
            result.write(f"# TYPE {family.name} {family.TYPE}\n")

            for key, value_set in fam_instance._get_value_sets():
                if value_set.timestamp is None:
                    timestamp_s = ''
                else:
                    timestamp_s = ' {0:d}'.format(value_set.timestamp)
                for suffix, value in value_set.values:
                    value_s = _floatToGoString(value)
                    result.write(f"{family.name}{suffix}{key} {value_s}{timestamp_s}\n")

        return result.getvalue()
